import sys
import gobject
import gtk
import gtk.gdk as gdk
import pango
import bonobo
import gosd
import gconfsync
import gc
import os.path
import gnome
import gnome_osd_conf
import warnings
import xml.dom.minidom as dom


bgcolor = "#000000"
fgcolor = "#80ff80"
gosd.BORDER_WIDTH=1
ANIMATION=True
ANIMATION_STEPS=10
ANIMATION_TIME=200

gnome.program_init("gnome-osd-server", gnome_osd_conf.VERSION)
prefs = gconfsync.GConfSync("/apps/gnome-osd",
                            dict(osd_halignment=dict(left=pango.ALIGN_LEFT,
                                                     right=pango.ALIGN_RIGHT,
                                                     center=pango.ALIGN_CENTER)))

# Note: interning atoms avoids X server roundtrips
_NET_CURRENT_DESKTOP = gtk.gdk.atom_intern("_NET_CURRENT_DESKTOP")
_NET_WORKAREA = gtk.gdk.atom_intern("_NET_WORKAREA")
def get_current_workarea():
    root = gtk.gdk.screen_get_default().get_root_window()
    prop = root.property_get(_NET_CURRENT_DESKTOP)
    if prop is None: return None
    propname, proptype, propvalue = prop
    current = propvalue[0]
    prop = root.property_get(_NET_WORKAREA)
    if prop is None: return None
    propname, proptype, propvalue = prop
    return propvalue[current*4 : (current + 1)*4]


class OSDMessage(object):

    class Message(object):
        __slots__ = ('text', 'timeout')

    class Animation(object):
        def __init__(self, win, x0, y0, xf, yf, callback, arg, enable_animation):
            self.callback = callback
            self.callback_arg = arg

            if not enable_animation:
                win.move(xf, yf)
                win.show()
                self.__animation_finished()
                return
                
            self.x = float(x0)
            self.y = float(y0)
            self.dx = (xf - x0) / ANIMATION_STEPS
            self.dy = (yf - y0) / ANIMATION_STEPS
            self.win = win
            self.counter = ANIMATION_STEPS
            self.timeout = gobject.timeout_add(ANIMATION_TIME // ANIMATION_STEPS,
                                               self.__timeout_cb)
            win.move(x0, y0)
            win.show()

        def __timeout_cb(self):
            self.x += self.dx
            self.y += self.dy
            self.win.move(int(self.x), int(self.y))
            self.counter -= 1
            if not self.counter:
                self.__animation_finished()
                return False
            return True

        def __animation_finished(self):
            self.timeout = 0
            if self.callback is not None:
                if self.callback_arg is None:
                    self.callback()
                else:
                    self.callback(self.callback_arg)
            self.win = None
            self.callback = None
            self.callback_arg = None

    
    def __init__(self, id, inherit=None):
        self.win = None
        self.hide_timeout = None
        self.message_queue = []
        self.__show_lock = False
        self.animation_vector = None
        self.inherit = inherit
        self.props = {}
        self.id = id
        self.occupied_area = None

    def getprop(self, name):
        try:
            return self.props[name]
        except KeyError:
            if self.inherit is None:
                return prefs[name]
            else:
                return self.inherit.getprop(name)

    def show(self, text, timeout=None):
        msg = self.Message()
        msg.text = text
        if timeout is None:
            msg.timeout = self.getprop('hide_timeout')
        else:
            msg.timeout = timeout
        self.message_queue.append(msg)
        self.__show_next()

    def __show_next(self):
        if self.__show_lock:
            return
        self.__show_lock = True
        if self.win is not None:
            self.hide(self.__show_next1)
        else:
            self.__show_next1()
            
    def __show_next1(self):
        halign = self.getprop('osd_halignment')
        vpos = self.getprop('osd_vposition')
        msg = self.message_queue.pop(0)

        if self.getprop('avoid_panels'):
            workarea = get_current_workarea()
            if workarea is not None:
                desk_x, desk_y, desk_w, desk_h = workarea
            else:
                warnings.warn("get_current_workarea failed, somehow")
                desk_x, desk_y = 0, 0
                desk_w = gdk.screen_width()
                desk_h = gdk.screen_height()
        else:
            desk_x, desk_y = 0, 0
            desk_w = gdk.screen_width()
            desk_h = gdk.screen_height()
        
        w = gosd.osd(msg.text, bgcolor, fgcolor,
                     pango.FontDescription(self.getprop('osd_font')),
                     use_markup=True,
                     alignment=halign,
                     fake_translucent_bg=self.getprop("osd_fake_translucent_bg"),
                     drop_shadow=self.getprop("drop_shadow"),
                     max_width=desk_w)
        if self.getprop('hide_on_hover'):
            w.connect("motion-notify-event", self.__osd_motion_notify)
        if self.getprop('enable_sound'):
            gnome.sound_play(self.__get_sound_sample_file())

        y, x, y_final = self.reserve_display_area(vpos, halign, w,
                                                  desk_x, desk_y, desk_w, desk_h)
        self.win = w
        self.Animation(w, x, y, x, y_final,
                       self.__show_next2, msg,
                       self.getprop('animations'))
        self.animation_vector = (x, y, x, y_final)

    def __show_next2(self, msg):
        self.hide_timeout = gobject.timeout_add(msg.timeout, self.hide)
        self.__show_lock = False
        if self.message_queue:
            self.__show_next()

    def hide(self, callback=None, cb_args=None):
        self.occupied_area = None
        if self.win is None:
            if callback is not None:
                if cb_args is None:
                    callback()
                else:
                    callback(cb_args)
            return
        
        if not self.hide_timeout:
            return
        gobject.source_remove(self.hide_timeout)
        self.hide_timeout = None
        xi, yi, xf, yf = self.animation_vector
        self.Animation(self.win, xf, yf, xi, yi,
                       self.__hide1, (callback, cb_args),
                       self.getprop('animations'))

    def __hide1(self, data):
        callback, cb_args = data
        self.win.destroy()
        self.win = None
        gc.collect()
        if callback is not None:
            if cb_args is None:
                callback()
            else:
                callback(cb_args)

    def __osd_motion_notify(self, widget, event):
        self.hide()

    def __get_sound_sample_file(self):
        sound_sample = self.getprop('sound_sample')
        if sound_sample[0] == '/':
            return sound_sample
        else:
            return os.path.join(gnome_osd_conf.datadir, "sounds", sound_sample)

    instances = {}

    def get(cls, attrs={}):
        global prefs
        id_ = attrs.get('id', 'default')
        inherit = attrs.get('inherit', 'default')
        try:
            osd_msg = cls.instances[id_]
        except KeyError:
            if inherit == 'default':
                inherit_osd = None
            else:
                inherit_osd = cls.instances[inherit]
            osd_msg = OSDMessage(id=id_, inherit=inherit_osd)
            cls.instances[id_] = osd_msg
        for key, str_val in attrs.iteritems():
            if key in ('id', 'inherit'): # special properties, no gconf schema
                continue
            if id_ == 'default':
                raise ValueError, "changing message properties"\
                      " for 'default' instance not allowed"
            val = prefs.parse_string(key, str_val)
            osd_msg.props[key] = val
        return osd_msg
    get = classmethod(get)

    def reserve_display_area(self, vpos, halign, w, desk_x, desk_y, desk_w, desk_h):
        if halign == pango.ALIGN_LEFT:
            x = 2 + desk_x
        elif halign == pango.ALIGN_RIGHT:
            x = desk_x + desk_w - w.width - 2
        else: # center
            x = desk_x + desk_w/2 - w.width/2

        occupied_areas = [inst.occupied_area for inst in self.instances.itervalues()
                          if not (inst is self or inst.occupied_area is None)]

        def intersection(area, x, y):
            x1, y1 = x, y
            x2, y2 = x + w.width, y + w.height
            ox1, oy1, ox2, oy2 = area
            result = ((x1 >= ox1 and x1 <= ox2
                       or x2 <= ox2 and x2 >= ox1
                       or x1 <= ox1 and x2 >= ox2)
                      and
                      (y1 >= oy1 and y1 <= oy2
                       or y2 <= oy2 and y2 >= oy1
                       or y1 <= oy1 and y2 >= oy2))
            return result
            
        if vpos == 'top':
            y = -w.height
            def sort_func(a, b):
                return cmp(a[3], b[3])
            occupied_areas.sort(sort_func)
            y_final = desk_y + 2
            for area in occupied_areas:
                if intersection(area, x, y_final):
                    y_final = area[3] + 2
                else:
                    break
        elif vpos == 'center':
            y = -w.height
            y_ideal = desk_y + desk_h//2  - w.height//2
            def sort_func(a, b):
                return cmp(abs(y_ideal - a), abs(y_ideal - b))
            y_final = y_ideal
            possible_targets = [y_ideal]
            for area in occupied_areas:
                possible_targets.append(area[3] + 2)
                possible_targets.append(area[1] - 2 - w.height)
            possible_targets.sort(sort_func)
            for target in possible_targets:
                for area in occupied_areas:
                    if intersection(area, x, target):
                        break
                else:
                    y_final = target
                    break
            else:
                y_final = None
            assert y_final is not None
        elif vpos == 'bottom':
            y = gdk.screen_height()
            y_final = desk_y + desk_h - w.height - 2
            def sort_func(a, b):
                return cmp(b[1], a[1])
            occupied_areas.sort(sort_func)
            for area in occupied_areas:
                if intersection(area, x, y_final):
                    y_final = area[1] - 2 - w.height
                else:
                    break
        else:
            raise ValueError, "invalid vposition '%s'" % vpos
        self.occupied_area = (x, y_final, x + w.width, y_final + w.height)
        return y, x, y_final
        

def message_quit_cb(app_client, message, app):
    print "received quit message"
    app.unref()
    bonobo.main_quit()

def message_show_cb(app_client, message, timeout=-1):
    try:
        attrs = {'id': 'default'}
        if timeout == -1:
            timeout = None
        osd_msg = OSDMessage.get(attrs)
        osd_msg.show(message, timeout)
        return ''
    except Exception, ex:
        print >> sys.stderr, ex
        return "%s: %s" % (str(ex.__class__), str(ex))

def message_parse(message):
    '''message_parse(msg) -> (attrs, pango_markup)'''
    doc = dom.parseString(message.decode("utf-8"))
    root = doc.documentElement
    if root.tagName != "message":
        raise RuntimeError("Protocol error: expecting <message>, got <%s>" % root.tagName)
    pango_markup = "".join(map(dom.Node.toxml, root.childNodes))
    attrs = {}
    for i in xrange(root.attributes.length):
        name = root.attributes.item(i).name
        attrs[name] = root.getAttribute(name)
    return attrs, pango_markup

def message_show_full_cb(app_client, message):
    try:
        attrs, pango_s = message_parse(message)
        osd_msg = OSDMessage.get(attrs)
        osd_msg.show(pango_s)
        return ''
    except Exception, ex:
        print >> sys.stderr, ex
        return "%s: %s" % (str(ex.__class__), str(ex))


def main():
    bonobo.activate()
    app = bonobo.Application("GNOME_OSD")
    def last_unref_cb(*args):
        print "Last Unref, quitting"
        bonobo.main_quit()
    bonobo.activation.active_server_register("OAFIID:GNOME_OSD", app.corba_objref())
    app.connect("message::quit", message_quit_cb, app)

    app.register_message("show",
 			 "Show a message in pango markup using default options",
 			 str, (str, int),
 			 message_show_cb)

    app.register_message("show-full",
			 "Show a message in Gnome OSD protocol format"
                         " (XML <message ...> ... </message>)",
			 str, (str,),
			 message_show_full_cb)

    bonobo.context_running_get().connect("last-unref", last_unref_cb)
    print "entering main loop..."
    bonobo.main()
